/* **************************************************************************

  SPACE.h, a library of general spatial and geometric utilities.
  Space.h contains jasort, an N-dimensional spatial sorting algorithm
of my own invention.

Copyright (C) Feb 19, 1999 John Conner Atkeson


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

For further information, contact:
John Conner Atkeson III
jcatkeson at gmail dot com
http://jca3.freeshell.org
3959 Persimmon Drive #103 Fairfax Va. 22031

************************************************************************** */

#ifndef space_h
#define space_h 1

#include <math.h>
#include "rosetta.h"

#define MaxDims 3  /* max dimensions. */
#define SortDims 2 /* dimensions to sort by. (XY) */
#define MaxNeigh 50  /* maximum number of neighbors, not used. */
#define Pi 3.141592653589
#define TwoPi Pi*2
// #define fudge 0.0000000000001/* to workaround divide-by-zero. */

/* some numeric types */
#ifndef maxint32 /* biggest possible long integer. */
  #define maxint32 0x7FFFFFFF
#endif
#ifndef sysint
  #define sysint int
#endif
#ifndef tin_bool
  #define tin_bool unsigned char
  #define true 0x1
  #define false 0x0
#endif
#ifndef byte
  #define byte unsigned char
#endif
#ifndef ulong
  #define ulong unsigned long
#endif
/* Whatever is the largest array-indexing integer that we can use on this system. */
#ifndef intdex
  #define intdex long
#endif
#ifndef true
  #define true 0x01
#endif
#ifndef false
  #define false 0x00
#endif
/* self-identity for preprocessor recognition. */
#define _spacelib 1

  /* N-dimensional bounding box. For more dimensions, just increase MaxDims. */
  typedef struct CubeT{
    double L[2][MaxDims]; /* limits of volume. */
    double Sz[MaxDims];   /* Wdt,Hgt,Dep of space; */
  }CubeT;
  typedef CubeT BoxType;  /* 2d version of cubetype, wastes Z. */

/* rubber sheet parameters for graphics/space. */
  typedef struct FrameT {  /* used for mapping any ortho surface */
    CubeT Wn,Vw;  /* onto any other. */
    double scale[MaxDims];
    double ShfScaleShf[MaxDims];
  } FrameT;
  
  /* 3d point type. */
  typedef struct XYZType *PtPtr;
  typedef struct XYZType{
    double Loc[MaxDims];/* xyz location */
    void *ref;/* generic extra values, usually NULL. */
  }XYZType;

/*
  Always index plist_class and VectT with "intdex" type, to keep
code upgradable to 64-bit.
*/
  /* Array of 3d points to sort. */
  typedef XYZType PRay[];
#if false
  typedef struct plist_class{
    intdex NumPoints;
	CubeT MinMax;  /* min and max xyz bounds for data set. */
    XYZType *P;
  }plist_class;
#endif
  /* ********************************************************************************************************************** */
  class plist_class{	    
  public:
	  intdex NumPoints;
	  CubeT MinMax;  /* min and max xyz bounds for data set. */
	  XYZType *P;
	  plist_class::plist_class(intdex NumElements)
	  /* ***********************************************************
	  *  Create a point list to size, all zeros.
	  *********************************************************** */
	  {
		  P=allocsafe(NumElements,XYZType);  
		  if(P==NULL){
			  printf(" Totally out of memory for index! \r\n");
			  printf("Press cntrl-c to avoid a computer crash. \r\n");
			  getchar();
		  }
		  NumPoints=NumElements;
		  blankrefs();
	  }
	  plist_class::~plist_class()
	  /* ***********************************************************
	  *  Dispose of a PList to free memory.
	  *********************************************************** */
	  {
		  freesafe(P);
		  P=NULL; NumPoints=0;
	  }/* plist_class::~plist_class */
	  void blankrefs()
	  /* ***********************************************************
	  *  Set all of the ref fields in a plist to NULL.
	  *********************************************************** */
	  { 
		  intdex pcnt;
		  for (pcnt=0;pcnt<NumPoints;pcnt++){
			  P[pcnt].ref=NULL;
		  }
	  }/* blankrefs */
  };/* plist_class */
  /* ********************************************************************************************************************** */

  /* Sort vector (sort index) type. */
/*  typedef sysint VRay; */
  typedef intdex VRay;
  typedef struct VectT{
    intdex NumPoints;
    plist_class *I_sort; /* the point-list that this vector index. */
    VRay *V;
  }VectT;
  
/*  typedef struct LogicType *LogicPtr; */
  typedef tin_bool LogicType(XYZType *Pt);
  
  typedef struct FunType *FunPtr;

/*
  bonded pair of a data array and its sort vector.
  pairs the data with a given Sort5x index.
  pairs the data with a given Sort index.
*/
typedef struct SortPairT{
  VectT *SVect;
  plist_class *PList;
} SortPairT;


/* Polygon/polyline type. */
typedef struct xyT{
  double xy[2];
}xyT;

typedef struct polyT{
  int ncorners;/* number of vertices. */
  xyT *vtex;
  tin_bool closed;/* open polyline or closed polygon */
}polyT;

/* public functions. */

void SortCube(CubeT *Cube);
void MoveCube(double movex,double movey,double movez,CubeT *Z);
void ExpandCube(double factor, CubeT *Z);
void OrthoFitCube(const CubeT *ABox,CubeT *BBox);
void SetCube(double x1,double y1,double z1,double x2,double y2,double z2,CubeT *Z); 
void SetCubeSz(CubeT *Z);
void SetBoxWH(CubeT *bx);
void SetBox(double x1,double y1,double x2,double y2,CubeT *bx);
void FlipCubeDim(CubeT *bx,int dimension);
void PolarizeND(const CubeT *ABox,CubeT *BBox,int dimension);
void OrthoFitBox(BoxType ABox,BoxType *BBox);
void OrthoFitND(const CubeT *ABox,CubeT *BBox,int dimension);
void BorderCube(CubeT *Z,double border);
void AreaFit2d(CubeT *Z,double newWdivH);
tin_bool InSpace(XYZType Point,CubeT *ThisBox,byte Dimensions);

VectT *sort_vect_create(intdex NumElements,bool fill);
VectT *FreeSortVect(VectT *Vec);
void FlipSortVect(VectT *Vec);

void jasort(plist_class *AllSurv2,VectT *XYSort2);
/* void TraverseZone(VectT *SVect,plist_class *PList,CubeT *Zone,intdex MinLoc,intdex MaxLoc,size_t Level,VectT *outray);*/
void TraverseZone(VectT *SVect,plist_class *PList,CubeT *Zone,VectT *outray);
#if 0
int GetNeighbors(plist_class *AllSurv2,VectT *XYSort2,intdex CenterDex,LogicType *Filter,VectT *Nbrs);
/* pythagorean closest. */
tin_bool FindClosest(contextT *context,intdex CenterDex,intdex *NbrDex,LogicType *Filter);
/* triangular closest. */
tin_bool FindThird(contextT *context,intdex Pt0,intdex Pt1,intdex *MinSubjIndex,LogicType *Filter);
#endif
void ResetFrame(FrameT *Fr);
void SetFrame(CubeT *Wn,CubeT *Vw,FrameT *Fr);
void ReproportionFrame(double x0,double y0,double z0,double x1,double y1,double z1,FrameT *Fr);

void Prj(double xloc,double yloc,const FrameT *F,double *Outx,double *Outy);
void DePrj(double xloc,double yloc,const FrameT *F,double *Inx,double *Iny); /* Descale from window. */
void Rotate(double OldX,double OldY,double Angle,double *NewX,double *NewY);
double Hypot(double Xdif,double Ydif);
double AngleSpread(double X1,double Y1,double XVert,double YVert,double X2,double Y2);

void create_blank_poly(polyT *p,int numofsides);
void realloc_poly(polyT *p,int numofsides);
void freepoly(polyT *p);
void poly_addvertex(polyT *p,double newx,double newy);
void RegPoly(double xorg,double yorg,double radius,int numofsides,polyT *p);
void get_poly_limits(polyT *p,CubeT *minmax);
tin_bool inside_poly(polyT *poly,double xloc,double yloc);
void get_points_inside_poly(VectT *xyvect, plist_class *xydata, polyT *poly, VectT *results);

#endif // ifndef space_h

